"use strict";
var ak_editor_prosemirror_1 = require("ak-editor-prosemirror");
var ak_editor_schema_1 = require("ak-editor-schema");
var ListsState = (function () {
    function ListsState(pm) {
        var _this = this;
        this.changeHandlers = [];
        // public state
        this.bulletListActive = false;
        this.bulletListDisabled = false;
        this.bulletListHidden = false;
        this.orderedListActive = false;
        this.orderedListDisabled = false;
        this.orderedListHidden = false;
        this.active = false;
        this.enabled = true;
        this.pm = pm;
        this.changeHandlers = [];
        var _a = pm.schema.nodes, bullet_list = _a.bullet_list, ordered_list = _a.ordered_list;
        this.bulletListHidden = !bullet_list;
        this.orderedListHidden = !ordered_list;
        this.wrapInBulletList = !!bullet_list ? ak_editor_prosemirror_1.commands.wrapInList(bullet_list) : noop;
        this.wrapInOrderedList = !!ordered_list ? ak_editor_prosemirror_1.commands.wrapInList(ordered_list) : noop;
        this.addKeymap(pm);
        pm.updateScheduler([
            pm.on.selectionChange,
            pm.on.change,
        ], function () { return _this.update(); });
    }
    ListsState.prototype.subscribe = function (cb) {
        this.changeHandlers.push(cb);
        cb(this);
    };
    ListsState.prototype.unsubscribe = function (cb) {
        this.changeHandlers = this.changeHandlers.filter(function (ch) { return ch !== cb; });
    };
    ListsState.prototype.toggleOrderedList = function () {
        var ordered_list = this.pm.schema.nodes.ordered_list;
        if (ordered_list) {
            this.toggleList(ordered_list);
        }
    };
    ListsState.prototype.toggleBulletList = function () {
        var bullet_list = this.pm.schema.nodes.bullet_list;
        if (bullet_list) {
            this.toggleList(bullet_list);
        }
    };
    ListsState.prototype.addKeymap = function (pm) {
        var _this = this;
        var list_item = pm.schema.nodes.list_item;
        pm.addKeymap(new ak_editor_prosemirror_1.Keymap({
            'Enter': function () { return ak_editor_prosemirror_1.commands.splitListItem(list_item)(pm); },
            'Mod-Shift-L': function () { return _this.toggleOrderedList(); },
            'Mod-Shift-B': function () { return _this.toggleBulletList(); }
        }));
    };
    /**
     * If the current selection is a list or part of a list, this method will untoggle
     * the list type for that selection.
     * In any other case it will try to apply the specified list type to the seletion,
     * either by converting existing lists to the new type or just apply the list type
     * if there's no list in the selection.
     */
    ListsState.prototype.toggleList = function (nodeType) {
        var _this = this;
        var groups = this.getGroupsInRange(this.pm.selection.$from, this.pm.selection.$to);
        var $from = groups[0].$from;
        var $to = groups[groups.length - 1].$to;
        this.pm.setSelection(new ak_editor_prosemirror_1.TextSelection($from, $to));
        var adjustedSelection = this.adjustSelection(this.pm.selection);
        var shouldUntoggle = this.isRangeOfType(adjustedSelection.$from, adjustedSelection.$to, nodeType);
        var rangeContainsList = this.rangeContainsList(adjustedSelection.$from, adjustedSelection.$to);
        var shouldConvertToType = !shouldUntoggle && (isListNode($from.node(1)) || rangeContainsList);
        groups.reverse(); // Quick-fix to keep positions
        groups.forEach(function (group) {
            _this.toggleListAtSelection(nodeType, group.$from, group.$to, shouldUntoggle, shouldConvertToType);
        });
    };
    ListsState.prototype.toggleListAtSelection = function (nodeType, $from, $to, shouldUntoggle, shouldConvertToType) {
        var pm = this.pm;
        pm.setSelection(new ak_editor_prosemirror_1.TextSelection($from, $to));
        var adjustedSelection = this.adjustSelection(pm.selection);
        var wrapInList = nodeType === pm.schema.nodes.bullet_list
            ? this.wrapInBulletList
            : this.wrapInOrderedList;
        if ($from === $to) {
            pm.setSelection(adjustedSelection);
            $from = pm.selection.$from;
            $to = pm.selection.$to;
        }
        if (shouldUntoggle) {
            if ($from.parent === $to.parent) {
                ak_editor_prosemirror_1.commands.lift(pm, true);
                return;
            }
            this.liftSelection(adjustedSelection.$from, adjustedSelection.$to).applyAndScroll();
            this.resetSelection();
        }
        else if (shouldConvertToType) {
            var tr = this.liftSelection(adjustedSelection.$from, adjustedSelection.$to).applyAndScroll();
            pm.setSelection(tr.selection);
            ak_editor_prosemirror_1.commands.wrapInList(nodeType)(tr.pm);
            if (this.shouldJoinUp(pm.selection, pm.doc, nodeType)) {
                ak_editor_prosemirror_1.commands.joinUp(pm, true);
            }
            if (this.shouldJoinDown(pm.selection, pm.doc, nodeType)) {
                ak_editor_prosemirror_1.commands.joinDown(pm, true);
            }
            this.resetSelection();
        }
        else {
            pm.setSelection(this.adjustSelection(pm.selection));
            ak_editor_prosemirror_1.commands.wrapInList(nodeType)(pm);
            if (this.shouldJoinUp(pm.selection, pm.doc, nodeType)) {
                ak_editor_prosemirror_1.commands.joinUp(pm, true);
            }
            if (this.shouldJoinDown(pm.selection, pm.doc, nodeType)) {
                /*
                 * joinDown expects the selection to be from the end of our last node to
                 * the beginning of the next. So we need to adjust our selection a bit.
                 * */
                pm.setSelection(new ak_editor_prosemirror_1.TextSelection(pm.selection.$to, pm.doc.resolve(pm.selection.$to.after(this.findAncestorPosition(pm.selection.$to).depth))));
                ak_editor_prosemirror_1.commands.joinDown(pm, true);
            }
            this.resetSelection();
        }
    };
    ListsState.prototype.findAncestorPosition = function (pos) {
        var nestableBlocks = ['blockquote', 'bullet_list', 'ordered_list'];
        if (pos.depth === 1) {
            return pos;
        }
        var node = pos.node(pos.depth);
        while (pos.depth >= 1) {
            pos = this.pm.doc.resolve(pos.before(pos.depth));
            node = pos.node(pos.depth);
            if (node && nestableBlocks.indexOf(node.type.name) !== -1) {
                break;
            }
        }
        return pos;
    };
    ListsState.prototype.hasCommonAncestor = function ($from, $to) {
        var current;
        var target;
        if ($from.depth > $to.depth) {
            current = this.findAncestorPosition($from);
            target = this.findAncestorPosition($to);
        }
        else {
            current = this.findAncestorPosition($to);
            target = this.findAncestorPosition($from);
        }
        while (current.depth > target.depth && current.depth > 1) {
            current = this.findAncestorPosition(current);
        }
        return current.node(current.depth) === target.node(target.depth);
    };
    ListsState.prototype.getGroupsInRange = function ($from, $to) {
        var groups = Array();
        var hasCommonAncestor = this.hasCommonAncestor($from, $to);
        var fromAncestor = this.findAncestorPosition($from);
        if (hasCommonAncestor || (fromAncestor.depth === 1 && isListNode($from.node(1)))) {
            groups.push({ $from: $from, $to: $to });
        }
        else {
            var current = $from;
            while (current.pos <= $to.pos) {
                var ancestorPos = this.findAncestorPosition(current);
                while (ancestorPos.depth > 1) {
                    ancestorPos = this.findAncestorPosition(ancestorPos);
                }
                var endPos = this.pm.doc.resolve(Math.min(ancestorPos.end(ancestorPos.depth) - 1, $to.pos));
                groups.push({
                    $from: current,
                    $to: endPos
                });
                current = this.pm.doc.resolve(Math.min(endPos.after(1) + 1, this.pm.doc.nodeSize - 2));
            }
        }
        return groups;
    };
    ListsState.prototype.update = function () {
        var _this = this;
        var pm = this.pm;
        var _a = pm.schema.nodes, bullet_list = _a.bullet_list, ordered_list = _a.ordered_list;
        var ancestorPosition = this.findAncestorPosition(pm.selection.$from);
        var rootNode = pm.selection instanceof ak_editor_prosemirror_1.NodeSelection
            ? pm.selection.node
            : ancestorPosition.node(ancestorPosition.depth);
        var dirty = false;
        var newBulletListActive = ak_editor_schema_1.isBulletListNode(rootNode);
        if (newBulletListActive !== this.bulletListActive) {
            this.bulletListActive = newBulletListActive;
            dirty = true;
        }
        var newOrderedListActive = ak_editor_schema_1.isOrderedListNode(rootNode);
        if (newOrderedListActive !== this.orderedListActive) {
            this.orderedListActive = newOrderedListActive;
            dirty = true;
        }
        var anyListActive = newBulletListActive || newOrderedListActive;
        var newBulletListDisabled = !(anyListActive || this.wrapInBulletList(pm, false));
        if (newBulletListDisabled !== this.bulletListDisabled) {
            this.bulletListDisabled = newBulletListDisabled;
            dirty = true;
        }
        var newOrderedListDisabled = !(anyListActive || this.wrapInOrderedList(pm, false));
        if (newOrderedListDisabled !== this.orderedListDisabled) {
            this.orderedListDisabled = newOrderedListDisabled;
            dirty = true;
        }
        if (dirty) {
            this.changeHandlers.forEach(function (cb) { return cb(_this); });
        }
    };
    /**
     * Returns all top-level ancestor-nodes between $from and $to
     */
    ListsState.prototype.ancestorNodesBetween = function ($from, $to) {
        var nodes = Array();
        var maxDepth = this.findAncestorPosition($from).depth;
        var current = this.pm.doc.resolve($from.start(maxDepth));
        while (current.pos <= $to.start($to.depth)) {
            var depth = Math.min(current.depth, maxDepth);
            var node = current.node(depth);
            if (node) {
                nodes.push(node);
            }
            var next = this.pm.doc.resolve(current.after(depth));
            if (next.start(depth) >= this.pm.doc.nodeSize - 2) {
                break;
            }
            if (next.depth !== current.depth) {
                next = this.pm.doc.resolve(next.pos + 2);
            }
            current = this.pm.doc.resolve(next.start(next.depth));
        }
        return nodes;
    };
    /**
     * Step through block-nodes between $from and $to and returns false if a node is
     * found that isn't of the specified type
     */
    ListsState.prototype.isRangeOfType = function ($from, $to, nodeType) {
        return this.ancestorNodesBetween($from, $to).filter(function (node) { return node.type !== nodeType; }).length === 0;
    };
    /**
     * Step through block-nodes between $from and $to and return true if a node is a
     * bullet_list or ordered_list
     */
    ListsState.prototype.rangeContainsList = function ($from, $to) {
        return this.ancestorNodesBetween($from, $to).some(isListNode);
    };
    /**
     * Takes a selection $from and $to and lift all text nodes from their parents to document-level
     */
    ListsState.prototype.liftSelection = function ($from, $to) {
        var tr = this.pm.tr;
        var startPos = $from.start($from.depth);
        var endPos = $to.end($to.depth);
        var target = Math.max(0, this.findAncestorPosition($from).depth - 1);
        tr.doc.nodesBetween(startPos, endPos, function (node, pos) {
            if (node.isText ||
                (node.isTextblock && !node.textContent) // Empty paragraph
            ) {
                var res = tr.doc.resolve(tr.map(pos));
                var sel = new ak_editor_prosemirror_1.NodeSelection(res);
                var range = sel.$from.blockRange(sel.$to);
                tr.lift(range, target);
            }
        });
        startPos = tr.map(startPos);
        endPos = tr.map(endPos);
        endPos = tr.doc.resolve(endPos).end(tr.doc.resolve(endPos).depth); // We want to select the entire node
        tr.setSelection(new ak_editor_prosemirror_1.TextSelection(tr.doc.resolve(startPos), tr.doc.resolve(endPos)));
        return tr;
    };
    /**
     * Determines if content inside a selection can be joined with the previous block.
     * We need this check since the built-in method for "joinUp" will join a ordered_list with bullet_list.
     */
    ListsState.prototype.shouldJoinUp = function (selection, doc, nodeType) {
        var res = doc.resolve(selection.$from.before(this.findAncestorPosition(selection.$from).depth));
        return res.nodeBefore && res.nodeBefore.type === nodeType;
    };
    /**
     * Determines if content inside a selection can be joined with the next block.
     * We need this check since the built-in method for "joinDown" will join a ordered_list with bullet_list.
     */
    ListsState.prototype.shouldJoinDown = function (selection, doc, nodeType) {
        var res = doc.resolve(selection.$to.after(this.findAncestorPosition(selection.$to).depth));
        return res.nodeAfter && res.nodeAfter.type === nodeType;
    };
    /**
     * Sometimes a selection in the editor can be slightly offset, for example:
     * it's possible for a selection to start or end at an empty node at the very end of
     * a line. This isn't obvious by looking at the editor and it's likely not what the
     * user intended - so we need to adjust the seletion a bit in scenarios like that.
     */
    ListsState.prototype.adjustSelection = function (selection) {
        var $from = selection.$from, $to = selection.$to;
        var isSameLine = $from.pos === $to.pos;
        if (isSameLine) {
            $from = this.pm.doc.resolve($from.start($from.depth));
            $to = this.pm.doc.resolve($from.end($from.depth));
        }
        var startPos = $from.pos;
        var endPos = $to.pos;
        if (isSameLine && startPos === this.pm.doc.nodeSize - 3) {
            return selection;
        }
        if ($from.nodeBefore) {
            if (!isSameLine) {
                startPos++;
                var node = this.pm.doc.nodeAt(startPos);
                while (!node || (node && !node.isText)) {
                    startPos++;
                    node = this.pm.doc.nodeAt(startPos);
                }
            }
            else if (!$from.nodeAfter) {
                startPos--;
                var node = this.pm.doc.nodeAt(startPos);
                while (!node || (node && !node.isText)) {
                    startPos--;
                    node = this.pm.doc.nodeAt(startPos);
                }
            }
        }
        if ($to.parentOffset) {
            endPos--;
        }
        else if ($to.nodeAfter && !($from.nodeAfter && isSameLine)) {
            endPos--;
            var node = this.pm.doc.nodeAt(endPos);
            while (node && !node.isText) {
                endPos--;
                node = this.pm.doc.nodeAt(endPos);
            }
        }
        if (!($from.parent && $from.parent.isTextblock && !$from.parent.textContent)) {
            var node = this.pm.doc.nodeAt(startPos);
            while (!node || (node && !node.isText)) {
                startPos++;
                node = this.pm.doc.nodeAt(startPos);
            }
        }
        if (!($to.parent && $to.parent.isTextblock && !$to.parent.textContent)) {
            var node = this.pm.doc.nodeAt(endPos);
            while (!node || (node && !node.isText)) {
                endPos--;
                node = this.pm.doc.nodeAt(endPos);
            }
        }
        if (endPos === startPos) {
            return new ak_editor_prosemirror_1.TextSelection(this.pm.doc.resolve(startPos));
        }
        return new ak_editor_prosemirror_1.TextSelection(this.pm.doc.resolve(startPos), this.pm.doc.resolve(endPos));
    };
    ListsState.prototype.resetSelection = function () {
        var newSelection = new ak_editor_prosemirror_1.TextSelection(this.pm.selection.$to);
        this.pm.setSelection(newSelection);
    };
    return ListsState;
}());
exports.ListsState = ListsState;
// IE11 + multiple prosemirror fix.
Object.defineProperty(ListsState, 'name', { value: 'ListsState' });
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = new ak_editor_prosemirror_1.Plugin(ListsState);
function isListNode(node) {
    return ak_editor_schema_1.isBulletListNode(node) || ak_editor_schema_1.isOrderedListNode(node);
}
var noop = function () {
    var args = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        args[_i] = arguments[_i];
    }
};
//# sourceMappingURL=index.js.map